package de.polygonal {
	import haxe.Int64;
	import de.polygonal._Printf.FloatType;
	import de.polygonal._Printf.IntegerType;
	import de.polygonal._Printf.FormatDataType;
	import haxe.ds.IntMap;
	import de.polygonal._Printf.FormatToken;
	public class Printf {
		static protected var dataTypeMap : haxe.ds.IntMap;
		static protected var formatIntFuncHash : haxe.ds.IntMap;
		static protected var formatFloatFuncHash : haxe.ds.IntMap;
		static protected var formatStringFuncHash : haxe.ds.IntMap;
		static protected var _initialized : Boolean = false;
		static protected function init() : void {
			de.polygonal.Printf.dataTypeMap = de.polygonal.Printf.makeDataTypeMap();
			de.polygonal.Printf.formatIntFuncHash = new haxe.ds.IntMap();
			de.polygonal.Printf.formatIntFuncHash.set(1,de.polygonal.Printf.formatSignedDecimal);
			de.polygonal.Printf.formatIntFuncHash.set(2,de.polygonal.Printf.formatUnsignedDecimal);
			de.polygonal.Printf.formatIntFuncHash.set(0,de.polygonal.Printf.formatCharacter);
			de.polygonal.Printf.formatIntFuncHash.set(4,de.polygonal.Printf.formatHexadecimal);
			de.polygonal.Printf.formatIntFuncHash.set(3,de.polygonal.Printf.formatOctal);
			de.polygonal.Printf.formatIntFuncHash.set(5,de.polygonal.Printf.formatBinary);
			de.polygonal.Printf.formatFloatFuncHash = new haxe.ds.IntMap();
			de.polygonal.Printf.formatFloatFuncHash.set(0,de.polygonal.Printf.formatNormalFloat);
			de.polygonal.Printf.formatFloatFuncHash.set(1,de.polygonal.Printf.formatScientific);
			de.polygonal.Printf.formatFloatFuncHash.set(2,de.polygonal.Printf.formatNaturalFloat);
			de.polygonal.Printf.formatStringFuncHash = new haxe.ds.IntMap();
			de.polygonal.Printf.formatStringFuncHash.set(2,de.polygonal.Printf.formatString);
		}
		
		static protected function makeDataTypeMap() : haxe.ds.IntMap {
			var hash : haxe.ds.IntMap = new haxe.ds.IntMap();
			hash.set(105,de.polygonal._Printf.FormatDataType.FmtInteger(de.polygonal._Printf.IntegerType.ISignedDecimal));
			hash.set(100,de.polygonal._Printf.FormatDataType.FmtInteger(de.polygonal._Printf.IntegerType.ISignedDecimal));
			hash.set(117,de.polygonal._Printf.FormatDataType.FmtInteger(de.polygonal._Printf.IntegerType.IUnsignedDecimal));
			hash.set(99,de.polygonal._Printf.FormatDataType.FmtInteger(de.polygonal._Printf.IntegerType.ICharacter));
			hash.set(120,de.polygonal._Printf.FormatDataType.FmtInteger(de.polygonal._Printf.IntegerType.IHex));
			hash.set(88,de.polygonal._Printf.FormatDataType.FmtInteger(de.polygonal._Printf.IntegerType.IHex));
			hash.set(111,de.polygonal._Printf.FormatDataType.FmtInteger(de.polygonal._Printf.IntegerType.IOctal));
			hash.set(98,de.polygonal._Printf.FormatDataType.FmtInteger(de.polygonal._Printf.IntegerType.IBin));
			hash.set(102,de.polygonal._Printf.FormatDataType.FmtFloat(de.polygonal._Printf.FloatType.FNormal));
			hash.set(101,de.polygonal._Printf.FormatDataType.FmtFloat(de.polygonal._Printf.FloatType.FScientific));
			hash.set(69,de.polygonal._Printf.FormatDataType.FmtFloat(de.polygonal._Printf.FloatType.FScientific));
			hash.set(103,de.polygonal._Printf.FormatDataType.FmtFloat(de.polygonal._Printf.FloatType.FNatural));
			hash.set(71,de.polygonal._Printf.FormatDataType.FmtFloat(de.polygonal._Printf.FloatType.FNatural));
			hash.set(115,de.polygonal._Printf.FormatDataType.FmtString);
			hash.set(112,de.polygonal._Printf.FormatDataType.FmtPointer);
			hash.set(110,de.polygonal._Printf.FormatDataType.FmtNothing);
			return hash;
		}
		
		static public function format(fmt : String,args : Array) : String {
			if(!de.polygonal.Printf._initialized) {
				de.polygonal.Printf._initialized = true;
				de.polygonal.Printf.init();
			};
			{
				var _g1 : int = 0;
				var _g : int = args.length;
				while(_g1 < _g) {
					var i : int = _g1++;
					if(args[i] == null) args[i] = "null";
				}
			};
			var output : String = "";
			var argIndex : int = 0;
			var tokens : Array = de.polygonal.Printf.tokenize(fmt);
			{
				var _g2 : int = 0;
				while(_g2 < tokens.length) {
					var token : de.polygonal._Printf.FormatToken = tokens[_g2];
					++_g2;
					switch(Type.enumIndex(token)) {
					case 3:
					throw "invalid format specifier";
					break;
					case 0:
					{
						var str : String = token.params[0];
						output += str;
					}
					break;
					case 2:
					{
						var name : String = token.params[0];
						{
							if(!Reflect.hasField(args[0],name)) throw "no field named " + name;
							output += Std.string(Reflect.field(args[0],name));
						}
					}
					break;
					case 1:
					{
						var tagArgs : * = token.params[1];
						var type : de.polygonal._Printf.FormatDataType = token.params[0];
						{
							if(tagArgs.width != null) tagArgs.width = tagArgs.width;
							else tagArgs.width = function() : int {
								var $r : int;
								var $t : * = args[argIndex++];
								if(Std._is($t,int)) (($t) as int);
								else throw "Class cast error";
								$r = $t;
								return $r;
							}();
							if(tagArgs.precision != null) tagArgs.precision = tagArgs.precision;
							else tagArgs.precision = function() : int {
								var $r2 : int;
								var $t : * = args[argIndex++];
								if(Std._is($t,int)) (($t) as int);
								else throw "Class cast error";
								$r2 = $t;
								return $r2;
							}();
							var value : * = args[argIndex++];
							var formatFunction : Function;
							switch(Type.enumIndex(type)) {
							case 1:
							{
								var floatType : de.polygonal._Printf.FloatType = type.params[0];
								formatFunction = de.polygonal.Printf.formatFloatFuncHash.get(Type.enumIndex(floatType));
							}
							break;
							case 0:
							{
								var integerType : de.polygonal._Printf.IntegerType = type.params[0];
								formatFunction = de.polygonal.Printf.formatIntFuncHash.get(Type.enumIndex(integerType));
							}
							break;
							case 2:
							formatFunction = de.polygonal.Printf.formatStringFuncHash.get(2);
							break;
							case 3:
							throw "specifier 'p' is not supported";
							break;
							case 4:
							throw "specifier 'n' is not supported";
							break;
							};
							output += formatFunction(value,tagArgs);
						}
					}
					break;
					}
				}
			};
			return output;
		}
		
		static protected function tokenize(fmt : String) : Array {
			var length : int = fmt.length;
			var lastStr : StringBuf = new StringBuf();
			var i : int = 0;
			var c : int = 0;
			var tokens : Array = new Array();
			while(i < length) {
				var c1 : int = de.polygonal.Printf.codeAt(fmt,i++);
				if(c1 == 37) {
					c1 = de.polygonal.Printf.codeAt(fmt,i++);
					if(c1 == 37) lastStr.addChar(c1);
					else {
						if(lastStr.toString().length > 0) {
							tokens.push(de.polygonal._Printf.FormatToken.BareString(lastStr.toString()));
							lastStr = new StringBuf();
						};
						var token : de.polygonal._Printf.FormatToken;
						if(c1 == 40) {
							var endPos : int = fmt.indexOf(")",i);
							if(endPos == -1) token = de.polygonal._Printf.FormatToken.Unknown("named param",i);
							else {
								var paramName : String = fmt.substr(i,endPos - i);
								i = endPos + 1;
								token = de.polygonal._Printf.FormatToken.Property(paramName);
							}
						}
						else {
							var params : * = { flags : int(0), pos : -1, width : -1, precision : -1};
							while(c1 == 45 || c1 == 43 || c1 == 35 || c1 == 48 || c1 == 32) {
								if(c1 == 45) params.flags |= 1;
								else if(c1 == 43) params.flags |= 2;
								else if(c1 == 35) params.flags |= 8;
								else if(c1 == 48) params.flags |= 16;
								else if(c1 == 32) params.flags |= 4;
								c1 = de.polygonal.Printf.codeAt(fmt,i++);
							};
							if((params.flags & 1) != 0 && (params.flags & 16) != 0) params.flags &= 268435439;
							if((params.flags & 4) != 0 && (params.flags & 2) != 0) params.flags &= 268435451;
							if(c1 == 42) {
								params.width = null;
								c1 = de.polygonal.Printf.codeAt(fmt,i++);
							}
							else if(de.polygonal.Printf.isDigit(c1)) {
								params.width = 0;
								while(de.polygonal.Printf.isDigit(c1)) {
									params.width = c1 - 48 + params.width * 10;
									c1 = de.polygonal.Printf.codeAt(fmt,i++);
								};
								if(c1 == 36) {
									params.pos = params.width - 1;
									params.width = -1;
									c1 = de.polygonal.Printf.codeAt(fmt,i++);
									if(c1 == 42) {
										params.width = null;
										c1 = de.polygonal.Printf.codeAt(fmt,i++);
									}
									else if(de.polygonal.Printf.isDigit(c1)) {
										params.width = 0;
										while(de.polygonal.Printf.isDigit(c1)) {
											params.width = c1 - 48 + params.width * 10;
											c1 = de.polygonal.Printf.codeAt(fmt,i++);
										}
									}
								}
							};
							if(c1 == 46) {
								c1 = de.polygonal.Printf.codeAt(fmt,i++);
								if(c1 == 42) {
									params.precision = null;
									c1 = de.polygonal.Printf.codeAt(fmt,i++);
								}
								else if(de.polygonal.Printf.isDigit(c1)) {
									params.precision = 0;
									while(de.polygonal.Printf.isDigit(c1)) {
										params.precision = c1 - 48 + params.precision * 10;
										c1 = de.polygonal.Printf.codeAt(fmt,i++);
									}
								}
								else params.precision = 0;
							};
							while(c1 == 104 || c1 == 108 || c1 == 76) {
								switch(c1) {
								case 104:
								params.flags |= 32;
								break;
								case 108:
								params.flags |= 128;
								break;
								case 76:
								params.flags |= 64;
								break;
								};
								c1 = de.polygonal.Printf.codeAt(fmt,i++);
							};
							if(c1 == 69 || c1 == 71 || c1 == 88) params.flags |= 256;
							var type : de.polygonal._Printf.FormatDataType = de.polygonal.Printf.dataTypeMap.get(c1);
							if(type == null) token = de.polygonal._Printf.FormatToken.Unknown(String.fromCharCode(c1),i);
							else token = de.polygonal._Printf.FormatToken.Tag(type,params);
						};
						tokens.push(token);
					}
				}
				else lastStr.addChar(c1);
			};
			if(lastStr.toString().length > 0) tokens.push(de.polygonal._Printf.FormatToken.BareString(lastStr.toString()));
			return tokens;
		}
		
		static protected function formatBinary(value : int,args : *) : String {
			var output : String = "";
			var flags : int = args.flags;
			var precision : * = args.precision;
			var width : * = args.width;
			if(precision == -1) precision = 1;
			if(value != 0) {
				if((flags & 32) != 0) value &= 65535;
				var i : int = value;
				do {
					output = ((((i & 1) > 0)?"1":"0")) + output;
					i >>>= 1;
				} while(i > 0);
				if(precision > 1) {
					if(precision > output.length) output = de.polygonal.Printf.lpad(output,"0",precision);
					if((flags & 8) != 0) output = "b" + output;
				}
			};
			if((flags & 1) != 0) {
				if(width > output.length) return de.polygonal.Printf.rpad(output," ",width);
				else return output;
			}
			else if(width > output.length) return de.polygonal.Printf.lpad(output,(((flags & 16) != 0)?"0":" "),width);
			else return output;
			return null;
		}
		
		static protected function formatOctal(value : int,args : *) : String {
			var output : String = "";
			var flags : int = args.flags;
			var precision : * = args.precision;
			var width : * = args.width;
			if(precision == -1) precision = 1;
			if(value != 0) {
				if((flags & 32) != 0) value &= 65535;
				output = de.polygonal.Printf.toOct(value);
				if((flags & 8) != 0) output = "0" + output;
				if(precision > 1 && output.length < precision) output = de.polygonal.Printf.lpad(output,"0",precision);
			};
			if((flags & 1) != 0) {
				if(width > output.length) return de.polygonal.Printf.rpad(output," ",width);
				else return output;
			}
			else if(width > output.length) return de.polygonal.Printf.lpad(output,(((flags & 16) != 0)?"0":" "),width);
			else return output;
			return null;
		}
		
		static protected function formatHexadecimal(value : int,args : *) : String {
			var output : String = "";
			var flags : int = args.flags;
			var precision : * = args.precision;
			var width : * = args.width;
			if(precision == -1) precision = 1;
			if(value != 0) {
				if((flags & 32) != 0) value &= 65535;
				output = de.polygonal.Printf.toHex(value);
				if(precision > 1 && output.length < precision) output = de.polygonal.Printf.lpad(output,"0",precision);
				if((flags & 8) != 0 && value != 0) output = "0x" + output;
				if((flags & 256) != 0) output = output.toUpperCase();
				else output = output.toLowerCase();
			};
			if((flags & 1) != 0) {
				if(width > output.length) return de.polygonal.Printf.rpad(output," ",width);
				else return output;
			}
			else if(width > output.length) return de.polygonal.Printf.lpad(output,(((flags & 16) != 0)?"0":" "),width);
			else return output;
			return null;
		}
		
		static protected function formatUnsignedDecimal(value : int,args : *) : String {
			var output : String;
			var precision : * = args.precision;
			if(value >= 0) output = de.polygonal.Printf.formatSignedDecimal(value,args);
			else {
				var x : haxe.Int64 = haxe.Int64.make(0,value);
				output = haxe.Int64.toStr(x);
				if(precision > 1 && output.length < precision) output = de.polygonal.Printf.lpad(output,"0",precision);
				output = de.polygonal.Printf.padNumber(output,value,args.flags,args.width);
			};
			return output;
		}
		
		static protected function formatNaturalFloat(value : Number,args : *) : String {
			args.precision = 0;
			var formatedFloat : String = de.polygonal.Printf.formatNormalFloat(value,args);
			var formatedScientific : String = de.polygonal.Printf.formatScientific(value,args);
			if((args.flags & 8) != 0) {
				if(formatedFloat.indexOf(".") != -1) {
					var pos : int = formatedFloat.length - 1;
					while(de.polygonal.Printf.codeAt(formatedFloat,pos) == 48) pos--;
					formatedFloat = formatedFloat.substr(0,pos);
				}
			};
			if(formatedFloat.length <= formatedScientific.length) return formatedFloat;
			else return formatedScientific;
			return null;
		}
		
		static protected function formatScientific(value : Number,args : *) : String {
			var output : String = "";
			var flags : int = args.flags;
			var precision : * = args.precision;
			if(precision == -1) precision = 6;
			var sign : int;
			var exponent : int;
			if(value == 0) {
				sign = 0;
				exponent = 0;
				output += "0";
				if(precision > 0) {
					output += ".";
					{
						var _g : int = 0;
						while(_g < int(precision)) {
							var i : int = _g++;
							output += "0";
						}
					}
				}
			}
			else {
				if(value > 0.) sign = 1;
				else if(value < 0.) sign = -1;
				else sign = 0;
				value = Math.abs(value);
				exponent = Math.floor(Math.log(value) / 2.302585092994046);
				value = value / Math.pow(10,exponent);
				var p : Number = Math.pow(0.1,precision);
				value = de.polygonal.Printf.roundTo(value,p);
			};
			if(sign < 0) output += "-";
			else if((flags & 2) != 0) output += "+";
			else output += "";
			if(value != 0) output += de.polygonal.Printf.rpad(de.polygonal.Printf.str(value).substr(0,precision + 2),"0",precision + 2);
			if((flags & 256) != 0) output += "E";
			else output += "e";
			if(exponent >= 0) output += "+";
			else output += "-";
			if(exponent < 10) output += "00";
			else if(exponent < 100) output += "0";
			output += de.polygonal.Printf.str(de.polygonal.Printf.iabs(exponent));
			return output;
		}
		
		static protected function formatSignedDecimal(value : int,args : *) : String {
			var output : String;
			var flags : int = args.flags;
			var precision : * = args.precision;
			var width : * = args.width;
			if(precision == 0 && value == 0) output = "";
			else {
				if((flags & 32) != 0) value &= 65535;
				output = de.polygonal.Printf.str(de.polygonal.Printf.iabs(value));
				if(precision > 1 && output.length < precision) output = de.polygonal.Printf.lpad(output,"0",precision);
				if((flags & 16) != 0) output = de.polygonal.Printf.lpad(output,"0",((value < 0)?width - 1:width));
				if(value < 0) output = "-" + output;
			};
			if(value >= 0) {
				if((flags & 2) != 0) output = "+" + output;
				else if((flags & 4) != 0) output = " " + output;
			};
			if((flags & 1) != 0) output = de.polygonal.Printf.rpad(output," ",args.width);
			else output = de.polygonal.Printf.lpad(output," ",args.width);
			return output;
		}
		
		static protected function formatString(x : String,args : *) : String {
			var output : String = x;
			var precision : * = args.precision;
			var width : * = args.width;
			if(precision > 0) output = x.substr(0,precision);
			var k : int = output.length;
			if(width > 0 && k < width) {
				if((args.flags & 1) != 0) output = de.polygonal.Printf.rpad(output," ",width);
				else output = de.polygonal.Printf.lpad(output," ",width);
			};
			return output;
		}
		
		static protected function formatNormalFloat(value : Number,args : *) : String {
			var output : String;
			var flags : int = args.flags;
			var precision : * = args.precision;
			var width : * = args.width;
			if(precision == -1) precision = 6;
			if(precision == 0) {
				output = de.polygonal.Printf.str(de.polygonal.Printf.iabs(Math.round(value)));
				if((flags & 8) != 0) output += ".";
			}
			else {
				value = de.polygonal.Printf.roundTo(value,Math.pow(.1,precision));
				var decimalPlaces : * = precision;
				if(Math["isNaN"](value)) output = "NaN";
				else {
					var t : int = Std._int(Math.pow(10,decimalPlaces));
					output = de.polygonal.Printf.str(Std._int(value * t) / t);
					var i : int = output.indexOf(".");
					if(i != -1) {
						var _g : int = output.substr(i + 1).length;
						while(_g < int(decimalPlaces)) {
							var i1 : int = _g++;
							output += "0";
						}
					}
					else {
						output += ".";
						{
							var _g1 : int = 0;
							while(_g1 < int(decimalPlaces)) {
								var i2 : int = _g1++;
								output += "0";
							}
						}
					}
				}
			};
			if((flags & 2) != 0 && value >= 0) output = "+" + output;
			else if((flags & 4) != 0 && value >= 0) output = " " + output;
			if((flags & 16) != 0) output = de.polygonal.Printf.lpad(output,"0",((value < 0)?width - 1:width));
			if((flags & 1) != 0) output = de.polygonal.Printf.rpad(output," ",width);
			else output = de.polygonal.Printf.lpad(output," ",width);
			return output;
		}
		
		static protected function formatCharacter(x : int,args : *) : String {
			var output : String = String.fromCharCode(x);
			if(args.width > 1) {
				if((args.flags & 1) != 0) output = de.polygonal.Printf.rpad(output," ",args.width);
				else output = de.polygonal.Printf.lpad(output," ",args.width);
			};
			return output;
		}
		
		static protected function padNumber(x : String,n : Number,flags : int,width : int) : String {
			var k : int = x.length;
			if(width > 0 && k < width) {
				if((flags & 1) != 0) x = de.polygonal.Printf.rpad(x," ",width);
				else if(n >= 0) x = de.polygonal.Printf.lpad(x,(((flags & 16) != 0)?"0":" "),width);
				else if((flags & 16) != 0) x = "-" + de.polygonal.Printf.lpad(x.substr(1),"0",width);
				else x = de.polygonal.Printf.lpad(x," ",width);
			};
			return x;
		}
		
		static protected function lpad(s : String,c : String,l : int) : String {
			if(c.length <= 0) throw "c.length <= 0";
			while(s.length < l) s = c + s;
			return s;
		}
		
		static protected function rpad(s : String,c : String,l : int) : String {
			if(c.length <= 0) throw "c.length <= 0";
			while(s.length < l) s = s + c;
			return s;
		}
		
		static protected function toHex(x : int) : String {
			var n : uint = x;
			var s : String = x.toString(16);
			s = s.toUpperCase();
			return s;
		}
		
		static protected function toOct(x : int) : String {
			var s : String = "";
			var t : int = x;
			do {
				s = (t & 7) + s;
				t >>>= 3;
			} while(t > 0);
			return s;
		}
		
		static protected function iabs(x : int) : int {
			return Std._int(Math.abs(x));
		}
		
		static protected function str(x : *) : String {
			return Std.string(x);
		}
		
		static protected function codeAt(x : String,i : int) : int {
			return StringTools.fastCodeAt(x,i);
		}
		
		static protected function isDigit(x : int) : Boolean {
			return x >= 48 && x <= 57;
		}
		
		static protected function roundTo(x : Number,y : Number) : Number {
			return Math.round(x / y) * y;
		}
		
	}
}
