package de.polygonal.ds {
	import de.polygonal.ds.DLLNode;
	import de.polygonal.ds.Itr;
	import de.polygonal.ds.Cloneable;
	import de.polygonal.ds.DLLIterator;
	import de.polygonal.ds.ArrayUtil;
	import de.polygonal.ds.HashKey;
	import de.polygonal.ds.Comparable;
	import de.polygonal.ds.CircularDLLIterator;
	import de.polygonal.ds.DA;
	import flash.Boot;
	import de.polygonal.ds.Collection;
	public class DLL implements de.polygonal.ds.Collection{
		public function DLL(reservedSize : int = 0,maxSize : int = -1) : void { if( !flash.Boot.skip_constructor ) {
			this.maxSize = -1;
			this._reservedSize = reservedSize;
			this._size = 0;
			this._poolSize = 0;
			this._circular = false;
			this._iterator = null;
			if(reservedSize > 0) this._headPool = this._tailPool = new de.polygonal.ds.DLLNode(null,this);
			this.head = this.tail = null;
			this.key = de.polygonal.ds.HashKey.next();
			this.reuseIterator = false;
		}}
		
		public var key : int;
		public var head : de.polygonal.ds.DLLNode;
		public var tail : de.polygonal.ds.DLLNode;
		public var maxSize : int;
		public var reuseIterator : Boolean;
		public var _size : int;
		protected var _reservedSize : int;
		protected var _poolSize : int;
		protected var _headPool : de.polygonal.ds.DLLNode;
		protected var _tailPool : de.polygonal.ds.DLLNode;
		public var _circular : Boolean;
		protected var _iterator : de.polygonal.ds.Itr;
		public function isCircular() : Boolean {
			return this._circular;
		}
		
		public function close() : void {
			if(this._circular) return;
			this._circular = true;
			if(this._valid(this.head)) {
				this.tail.next = this.head;
				this.head.prev = this.tail;
			}
		}
		
		public function open() : void {
			if(!this._circular) return;
			this._circular = false;
			if(this._valid(this.head)) {
				this.tail.next = null;
				this.head.prev = null;
			}
		}
		
		public function createNode(x : *) : de.polygonal.ds.DLLNode {
			return new de.polygonal.ds.DLLNode(x,this);
		}
		
		public function append(x : *) : de.polygonal.ds.DLLNode {
			var node : de.polygonal.ds.DLLNode = this._getNode(x);
			if(this._valid(this.tail)) {
				this.tail.next = node;
				node.prev = this.tail;
			}
			else this.head = node;
			this.tail = node;
			if(this._circular) {
				this.tail.next = this.head;
				this.head.prev = this.tail;
			};
			this._size++;
			return node;
		}
		
		public function appendNode(x : de.polygonal.ds.DLLNode) : void {
			if(this._valid(this.tail)) {
				this.tail.next = x;
				x.prev = this.tail;
			}
			else this.head = x;
			this.tail = x;
			if(this._circular) {
				this.tail.next = this.head;
				this.head.prev = this.tail;
			};
			this._size++;
		}
		
		public function prepend(x : *) : de.polygonal.ds.DLLNode {
			var node : de.polygonal.ds.DLLNode = this._getNode(x);
			node.next = this.head;
			if(this._valid(this.head)) this.head.prev = node;
			else this.tail = node;
			this.head = node;
			if(this._circular) {
				this.tail.next = this.head;
				this.head.prev = this.tail;
			};
			this._size++;
			return node;
		}
		
		public function prependNode(x : de.polygonal.ds.DLLNode) : void {
			x.next = this.head;
			if(this._valid(this.head)) this.head.prev = x;
			else this.tail = x;
			this.head = x;
			if(this._circular) {
				this.tail.next = this.head;
				this.head.prev = this.tail;
			};
			this._size++;
		}
		
		public function insertAfter(node : de.polygonal.ds.DLLNode,x : *) : de.polygonal.ds.DLLNode {
			var t : de.polygonal.ds.DLLNode = this._getNode(x);
			this.__insertAfter(node,t);
			if(node == this.tail) {
				this.tail = t;
				if(this._circular) this.tail.next = this.head;
			};
			this._size++;
			return t;
		}
		
		public function insertBefore(node : de.polygonal.ds.DLLNode,x : *) : de.polygonal.ds.DLLNode {
			var t : de.polygonal.ds.DLLNode = this._getNode(x);
			this.__insertBefore(node,t);
			if(node == this.head) {
				this.head = t;
				if(this._circular) this.head.prev = this.tail;
			};
			this._size++;
			return t;
		}
		
		public function unlink(node : de.polygonal.ds.DLLNode) : de.polygonal.ds.DLLNode {
			var hook : de.polygonal.ds.DLLNode = node.next;
			if(node == this.head) {
				this.head = this.head.next;
				if(this._circular) {
					if(this.head == this.tail) this.head = null;
					else this.tail.next = this.head;
				};
				if(this.head == null) this.tail = null;
			}
			else if(node == this.tail) {
				this.tail = this.tail.prev;
				if(this._circular) this.head.prev = this.tail;
				if(this.tail == null) this.head = null;
			};
			this.__unlink(node);
			this._putNode(node);
			this._size--;
			return hook;
		}
		
		public function getNodeAt(i : int) : de.polygonal.ds.DLLNode {
			var node : de.polygonal.ds.DLLNode = this.head;
			{
				var _g : int = 0;
				while(_g < i) {
					var j : int = _g++;
					node = node.next;
				}
			};
			return node;
		}
		
		public function removeHead() : * {
			var node : de.polygonal.ds.DLLNode = this.head;
			if(this.head == this.tail) this.head = this.tail = null;
			else {
				this.head = this.head.next;
				node.next = null;
				if(this._circular) {
					this.head.prev = this.tail;
					this.tail.next = this.head;
				}
				else this.head.prev = null;
			};
			this._size--;
			return this._putNode(node);
		}
		
		public function removeTail() : * {
			var node : de.polygonal.ds.DLLNode = this.tail;
			if(this.head == this.tail) this.head = this.tail = null;
			else {
				this.tail = this.tail.prev;
				node.prev = null;
				if(this._circular) {
					this.tail.next = this.head;
					this.head.prev = this.tail;
				}
				else this.tail.next = null;
			};
			this._size--;
			return this._putNode(node);
		}
		
		public function shiftUp() : void {
			if(this._size > 1) {
				var t : de.polygonal.ds.DLLNode = this.head;
				if(this.head.next == this.tail) {
					this.head = this.tail;
					this.head.prev = null;
					this.tail = t;
					this.tail.next = null;
					this.head.next = this.tail;
					this.tail.prev = this.head;
				}
				else {
					this.head = this.head.next;
					this.head.prev = null;
					this.tail.next = t;
					t.next = null;
					t.prev = this.tail;
					this.tail = t;
				};
				if(this._circular) {
					this.tail.next = this.head;
					this.head.prev = this.tail;
				}
			}
		}
		
		public function popDown() : void {
			if(this._size > 1) {
				var t : de.polygonal.ds.DLLNode = this.tail;
				if(this.tail.prev == this.head) {
					this.tail = this.head;
					this.tail.next = null;
					this.head = t;
					this.head.prev = null;
					this.head.next = this.tail;
					this.tail.prev = this.head;
				}
				else {
					this.tail = this.tail.prev;
					this.tail.next = null;
					this.head.prev = t;
					t.prev = null;
					t.next = this.head;
					this.head = t;
				};
				if(this._circular) {
					this.tail.next = this.head;
					this.head.prev = this.tail;
				}
			}
		}
		
		public function nodeOf(x : *,from : de.polygonal.ds.DLLNode = null) : de.polygonal.ds.DLLNode {
			var node : de.polygonal.ds.DLLNode;
			if(from == null) node = this.head;
			else node = from;
			if(this._circular) {
				while(node != this.tail) {
					if(node.val == x) return node;
					node = node.next;
				};
				if(node.val == x) return node;
			}
			else while(this._valid(node)) {
				if(node.val == x) return node;
				node = node.next;
			};
			return null;
		}
		
		public function lastNodeOf(x : *,from : de.polygonal.ds.DLLNode = null) : de.polygonal.ds.DLLNode {
			var node : de.polygonal.ds.DLLNode;
			if(from == null) node = this.tail;
			else node = from;
			if(this._circular) {
				while(node != this.head) {
					if(node.val == x) return node;
					node = node.prev;
				};
				if(node.val == x) return node;
			}
			else while(this._valid(node)) {
				if(node.val == x) return node;
				node = node.prev;
			};
			return null;
		}
		
		public function sort(compare : Function,useInsertionSort : Boolean = false) : void {
			if(this._size > 1) {
				if(this._circular) {
					this.tail.next = null;
					this.head.prev = null;
				};
				if(compare == null) if(useInsertionSort) this.head = this._insertionSortComparable(this.head);
				else this.head = this._mergeSortComparable(this.head);
				else if(useInsertionSort) this.head = this._insertionSort(this.head,compare);
				else this.head = this._mergeSort(this.head,compare);
				if(this._circular) {
					this.tail.next = this.head;
					this.head.prev = this.tail;
				}
			}
		}
		
		public function merge(x : de.polygonal.ds.DLL) : void {
			if(this._valid(x.head)) {
				var node : de.polygonal.ds.DLLNode = x.head;
				{
					var _g1 : int = 0;
					var _g : int = x.size();
					while(_g1 < _g) {
						var i : int = _g1++;
						this.__list(node,this);
						node = node.next;
					}
				};
				if(this._valid(this.head)) {
					this.tail.next = x.head;
					x.head.prev = this.tail;
					this.tail = x.tail;
				}
				else {
					this.head = x.head;
					this.tail = x.tail;
				};
				this._size += x.size();
				if(this._circular) {
					this.tail.next = this.head;
					this.head.prev = this.tail;
				}
			}
		}
		
		public function concat(x : de.polygonal.ds.DLL) : de.polygonal.ds.DLL {
			var c : de.polygonal.ds.DLL = new de.polygonal.ds.DLL();
			var k : int = x.size();
			if(k > 0) {
				var node : de.polygonal.ds.DLLNode = x.tail;
				var t : de.polygonal.ds.DLLNode = c.tail = new de.polygonal.ds.DLLNode(node.val,c);
				node = node.prev;
				var i : int = k - 1;
				while(i-- > 0) {
					var copy : de.polygonal.ds.DLLNode = new de.polygonal.ds.DLLNode(node.val,c);
					copy.next = t;
					t.prev = copy;
					t = copy;
					node = node.prev;
				};
				c.head = t;
				c._size = k;
				if(this._size > 0) {
					var node1 : de.polygonal.ds.DLLNode = this.tail;
					var i1 : int = this._size;
					while(i1-- > 0) {
						var copy1 : de.polygonal.ds.DLLNode = new de.polygonal.ds.DLLNode(node1.val,c);
						copy1.next = t;
						t.prev = copy1;
						t = copy1;
						node1 = node1.prev;
					};
					c.head = t;
					c._size += this._size;
				}
			}
			else if(this._size > 0) {
				var node2 : de.polygonal.ds.DLLNode = this.tail;
				var t1 : de.polygonal.ds.DLLNode = c.tail = new de.polygonal.ds.DLLNode(node2.val,this);
				node2 = node2.prev;
				var i2 : int = this._size - 1;
				while(i2-- > 0) {
					var copy2 : de.polygonal.ds.DLLNode = new de.polygonal.ds.DLLNode(node2.val,this);
					copy2.next = t1;
					t1.prev = copy2;
					t1 = copy2;
					node2 = node2.prev;
				};
				c.head = t1;
				c._size = this._size;
			};
			return c;
		}
		
		public function reverse() : void {
			if(this._size <= 1) return;
			else if(this._size <= 3) {
				var t : * = this.head.val;
				this.head.val = this.tail.val;
				this.tail.val = t;
			}
			else {
				var head : de.polygonal.ds.DLLNode = this.head;
				var tail : de.polygonal.ds.DLLNode = this.tail;
				{
					var _g1 : int = 0;
					var _g : int = this._size >> 1;
					while(_g1 < _g) {
						var i : int = _g1++;
						var t1 : * = head.val;
						head.val = tail.val;
						tail.val = t1;
						head = head.next;
						tail = tail.prev;
					}
				}
			}
		}
		
		public function join(x : String) : String {
			var s : String = "";
			if(this._size > 0) {
				var node : de.polygonal.ds.DLLNode = this.head;
				{
					var _g1 : int = 0;
					var _g : int = this._size - 1;
					while(_g1 < _g) {
						var i : int = _g1++;
						s += Std.string(node.val) + x;
						node = node.next;
					}
				};
				s += Std.string(node.val);
			};
			return s;
		}
		
		public function assign(C : Class,args : Array = null,n : int = 0) : void {
			if(n > 0) {
			}
			else n = this.size();
			if(args == null) args = [];
			var node : de.polygonal.ds.DLLNode = this.head;
			{
				var _g : int = 0;
				while(_g < int(n)) {
					var i : int = _g++;
					node.val = Type.createInstance(C,args);
					node = node.next;
				}
			}
		}
		
		public function fill(x : *,args : Array = null,n : int = 0) : de.polygonal.ds.DLL {
			if(n > 0) {
			}
			else n = this.size();
			var node : de.polygonal.ds.DLLNode = this.head;
			{
				var _g : int = 0;
				while(_g < int(n)) {
					var i : int = _g++;
					node.val = x;
					node = node.next;
				}
			};
			return this;
		}
		
		public function shuffle(rval : de.polygonal.ds.DA = null) : void {
			var s : int = this._size;
			if(rval == null) {
				var m : * = Math;
				while(s > 1) {
					s--;
					var i : int = Std._int(m.random() * s);
					var node1 : de.polygonal.ds.DLLNode = this.head;
					{
						var _g : int = 0;
						while(_g < int(s)) {
							var j : int = _g++;
							node1 = node1.next;
						}
					};
					var t : * = node1.val;
					var node2 : de.polygonal.ds.DLLNode = this.head;
					{
						var _g1 : int = 0;
						while(_g1 < int(i)) {
							var j1 : int = _g1++;
							node2 = node2.next;
						}
					};
					node1.val = node2.val;
					node2.val = t;
				}
			}
			else {
				var j2 : int = 0;
				while(s > 1) {
					s--;
					var i1 : int = Std._int(rval.get(j2++) * s);
					var node11 : de.polygonal.ds.DLLNode = this.head;
					{
						var _g2 : int = 0;
						while(_g2 < int(s)) {
							var j3 : int = _g2++;
							node11 = node11.next;
						}
					};
					var t1 : * = node11.val;
					var node21 : de.polygonal.ds.DLLNode = this.head;
					{
						var _g3 : int = 0;
						while(_g3 < int(i1)) {
							var j4 : int = _g3++;
							node21 = node21.next;
						}
					};
					node11.val = node21.val;
					node21.val = t1;
				}
			};
			if(this._circular) {
				this.tail.next = this.head;
				this.head.prev = this.tail;
			}
		}
		
		public function toString() : String {
			var s : String = "{ DLL size: " + this.size() + ", circular: " + Std.string(this.isCircular()) + " }";
			if(this.isEmpty()) return s;
			s += "\n[ head \n";
			var node : de.polygonal.ds.DLLNode = this.head;
			{
				var _g1 : int = 0;
				var _g : int = this._size;
				while(_g1 < _g) {
					var i : int = _g1++;
					s += "  " + Std.string(node.val) + "\n";
					node = node.next;
				}
			};
			s += "] tail";
			return s;
		}
		
		public function free() : void {
			var node : de.polygonal.ds.DLLNode = this.head;
			{
				var _g1 : int = 0;
				var _g : int = this._size;
				while(_g1 < _g) {
					var i : int = _g1++;
					var next : de.polygonal.ds.DLLNode = node.next;
					node.free();
					node = next;
				}
			};
			this.head = this.tail = null;
			var node1 : de.polygonal.ds.DLLNode = this._headPool;
			while(this._valid(node1)) {
				var next1 : de.polygonal.ds.DLLNode = node1.next;
				node1.free();
				node1 = next1;
			};
			this._headPool = this._tailPool = null;
			this._iterator = null;
		}
		
		public function contains(x : *) : Boolean {
			var node : de.polygonal.ds.DLLNode = this.head;
			{
				var _g1 : int = 0;
				var _g : int = this._size;
				while(_g1 < _g) {
					var i : int = _g1++;
					if(node.val == x) return true;
					node = node.next;
				}
			};
			return false;
		}
		
		public function remove(x : *) : Boolean {
			var s : int = this.size();
			if(s == 0) return false;
			var node : de.polygonal.ds.DLLNode = this.head;
			while(this._valid(node)) if(node.val == x) node = this.unlink(node);
			else node = node.next;
			return this.size() < s;
		}
		
		public function clear(purge : Boolean = false) : void {
			if(purge || this._reservedSize > 0) {
				var node : de.polygonal.ds.DLLNode = this.head;
				{
					var _g1 : int = 0;
					var _g : int = this._size;
					while(_g1 < _g) {
						var i : int = _g1++;
						var next : de.polygonal.ds.DLLNode = node.next;
						node.prev = null;
						node.next = null;
						this._putNode(node);
						node = next;
					}
				}
			};
			this.head = this.tail = null;
			this._size = 0;
		}
		
		public function iterator() : de.polygonal.ds.Itr {
			if(this.reuseIterator) {
				if(this._iterator == null) {
					if(this._circular) return new de.polygonal.ds.CircularDLLIterator(this);
					else return new de.polygonal.ds.DLLIterator(this);
				}
				else this._iterator.reset();
				return this._iterator;
			}
			else if(this._circular) return new de.polygonal.ds.CircularDLLIterator(this);
			else return new de.polygonal.ds.DLLIterator(this);
			return null;
		}
		
		public function size() : int {
			return this._size;
		}
		
		public function isEmpty() : Boolean {
			return this._size == 0;
		}
		
		public function toArray() : Array {
			var a : Array = de.polygonal.ds.ArrayUtil.alloc(this.size());
			var node : de.polygonal.ds.DLLNode = this.head;
			{
				var _g1 : int = 0;
				var _g : int = this._size;
				while(_g1 < _g) {
					var i : int = _g1++;
					a[i] = node.val;
					node = node.next;
				}
			};
			return a;
		}
		
		public function toVector() : Vector.<*> {
			var a : Vector.<*> = new Vector.<*>(this.size());
			var node : de.polygonal.ds.DLLNode = this.head;
			{
				var _g1 : int = 0;
				var _g : int = this._size;
				while(_g1 < _g) {
					var i : int = _g1++;
					a[i] = node.val;
					node = node.next;
				}
			};
			return a;
		}
		
		public function clone(assign : Boolean = true,copier : Function = null) : de.polygonal.ds.Collection {
			if(this._size == 0) {
				var copy : de.polygonal.ds.DLL = new de.polygonal.ds.DLL(this._reservedSize,this.maxSize);
				if(this._circular) copy._circular = true;
				return copy;
			};
			var copy1 : de.polygonal.ds.DLL = new de.polygonal.ds.DLL();
			copy1._size = this._size;
			if(assign) {
				var srcNode : de.polygonal.ds.DLLNode = this.head;
				var dstNode : de.polygonal.ds.DLLNode = copy1.head = new de.polygonal.ds.DLLNode(this.head.val,copy1);
				if(this._size == 1) {
					copy1.tail = copy1.head;
					if(this._circular) copy1.tail.next = copy1.head;
					return copy1;
				};
				var dstNode0 : de.polygonal.ds.DLLNode;
				srcNode = srcNode.next;
				{
					var _g1 : int = 1;
					var _g : int = this._size - 1;
					while(_g1 < _g) {
						var i : int = _g1++;
						dstNode0 = dstNode;
						var srcNode0 : de.polygonal.ds.DLLNode = srcNode;
						dstNode = dstNode.next = new de.polygonal.ds.DLLNode(srcNode.val,copy1);
						dstNode.prev = dstNode0;
						srcNode0 = srcNode;
						srcNode = srcNode0.next;
					}
				};
				dstNode0 = dstNode;
				copy1.tail = dstNode.next = new de.polygonal.ds.DLLNode(srcNode.val,copy1);
				copy1.tail.prev = dstNode0;
			}
			else if(copier == null) {
				var srcNode1 : de.polygonal.ds.DLLNode = this.head;
				var c : de.polygonal.ds.Cloneable;
				c = (function($this:DLL) : de.polygonal.ds.Cloneable {
					var $r : de.polygonal.ds.Cloneable;
					var $t : * = $this.head.val;
					if(Std._is($t,de.polygonal.ds.Cloneable)) (($t) as de.polygonal.ds.Cloneable);
					else throw "Class cast error";
					$r = $t;
					return $r;
				}(this));
				var dstNode1 : de.polygonal.ds.DLLNode = copy1.head = new de.polygonal.ds.DLLNode(c.clone(),copy1);
				if(this._size == 1) {
					copy1.tail = copy1.head;
					if(this._circular) copy1.tail.next = copy1.head;
					return copy1;
				};
				var dstNode01 : de.polygonal.ds.DLLNode;
				srcNode1 = srcNode1.next;
				{
					var _g11 : int = 1;
					var _g2 : int = this._size - 1;
					while(_g11 < _g2) {
						var i1 : int = _g11++;
						dstNode01 = dstNode1;
						var srcNode01 : de.polygonal.ds.DLLNode = srcNode1;
						c = (function($this:DLL) : de.polygonal.ds.Cloneable {
							var $r2 : de.polygonal.ds.Cloneable;
							var $t : * = srcNode1.val;
							if(Std._is($t,de.polygonal.ds.Cloneable)) (($t) as de.polygonal.ds.Cloneable);
							else throw "Class cast error";
							$r2 = $t;
							return $r2;
						}(this));
						dstNode1 = dstNode1.next = new de.polygonal.ds.DLLNode(c.clone(),copy1);
						dstNode1.prev = dstNode01;
						srcNode01 = srcNode1;
						srcNode1 = srcNode01.next;
					}
				};
				c = (function($this:DLL) : de.polygonal.ds.Cloneable {
					var $r3 : de.polygonal.ds.Cloneable;
					var $t : * = srcNode1.val;
					if(Std._is($t,de.polygonal.ds.Cloneable)) (($t) as de.polygonal.ds.Cloneable);
					else throw "Class cast error";
					$r3 = $t;
					return $r3;
				}(this));
				dstNode01 = dstNode1;
				copy1.tail = dstNode1.next = new de.polygonal.ds.DLLNode(c.clone(),copy1);
				copy1.tail.prev = dstNode01;
			}
			else {
				var srcNode2 : de.polygonal.ds.DLLNode = this.head;
				var dstNode2 : de.polygonal.ds.DLLNode = copy1.head = new de.polygonal.ds.DLLNode(copier(this.head.val),copy1);
				if(this._size == 1) {
					copy1.tail = copy1.head;
					if(this._circular) copy1.tail.next = copy1.head;
					return copy1;
				};
				var dstNode02 : de.polygonal.ds.DLLNode;
				srcNode2 = srcNode2.next;
				{
					var _g12 : int = 1;
					var _g3 : int = this._size - 1;
					while(_g12 < _g3) {
						var i2 : int = _g12++;
						dstNode02 = dstNode2;
						var srcNode02 : de.polygonal.ds.DLLNode = srcNode2;
						dstNode2 = dstNode2.next = new de.polygonal.ds.DLLNode(copier(srcNode2.val),copy1);
						dstNode2.prev = dstNode02;
						srcNode02 = srcNode2;
						srcNode2 = srcNode02.next;
					}
				};
				dstNode02 = dstNode2;
				copy1.tail = dstNode2.next = new de.polygonal.ds.DLLNode(copier(srcNode2.val),copy1);
				copy1.tail.prev = dstNode02;
			};
			if(this._circular) copy1.tail.next = copy1.head;
			return copy1;
		}
		
		protected function _mergeSortComparable(node : de.polygonal.ds.DLLNode) : de.polygonal.ds.DLLNode {
			var h : de.polygonal.ds.DLLNode = node;
			var p : de.polygonal.ds.DLLNode;
			var q : de.polygonal.ds.DLLNode;
			var e : de.polygonal.ds.DLLNode;
			var tail : de.polygonal.ds.DLLNode = null;
			var insize : int = 1;
			var nmerges : int;
			var psize : int;
			var qsize : int;
			var i : *;
			while(true) {
				p = h;
				h = tail = null;
				nmerges = 0;
				while(this._valid(p)) {
					nmerges++;
					psize = 0;
					q = p;
					{
						var _g : int = 0;
						while(_g < int(insize)) {
							var i1 : int = _g++;
							psize++;
							q = q.next;
							if(q == null) break;
						}
					};
					qsize = insize;
					while(psize > 0 || qsize > 0 && this._valid(q)) {
						if(psize == 0) {
							e = q;
							q = q.next;
							qsize--;
						}
						else if(qsize == 0 || q == null) {
							e = p;
							p = p.next;
							psize--;
						}
						else if(((function($this:DLL) : de.polygonal.ds.Comparable {
							var $r : de.polygonal.ds.Comparable;
							var $t : * = p.val;
							if(Std._is($t,de.polygonal.ds.Comparable)) (($t) as de.polygonal.ds.Comparable);
							else throw "Class cast error";
							$r = $t;
							return $r;
						}(this))).compare(q.val) >= 0) {
							e = p;
							p = p.next;
							psize--;
						}
						else {
							e = q;
							q = q.next;
							qsize--;
						};
						if(this._valid(tail)) tail.next = e;
						else h = e;
						e.prev = tail;
						tail = e;
					};
					p = q;
				};
				tail.next = null;
				if(nmerges <= 1) break;
				insize <<= 1;
			};
			h.prev = null;
			this.tail = tail;
			return h;
		}
		
		protected function _mergeSort(node : de.polygonal.ds.DLLNode,cmp : Function) : de.polygonal.ds.DLLNode {
			var h : de.polygonal.ds.DLLNode = node;
			var p : de.polygonal.ds.DLLNode;
			var q : de.polygonal.ds.DLLNode;
			var e : de.polygonal.ds.DLLNode;
			var tail : de.polygonal.ds.DLLNode = null;
			var insize : int = 1;
			var nmerges : int;
			var psize : int;
			var qsize : int;
			var i : *;
			while(true) {
				p = h;
				h = tail = null;
				nmerges = 0;
				while(this._valid(p)) {
					nmerges++;
					psize = 0;
					q = p;
					{
						var _g : int = 0;
						while(_g < int(insize)) {
							var i1 : int = _g++;
							psize++;
							q = q.next;
							if(q == null) break;
						}
					};
					qsize = insize;
					while(psize > 0 || qsize > 0 && this._valid(q)) {
						if(psize == 0) {
							e = q;
							q = q.next;
							qsize--;
						}
						else if(qsize == 0 || q == null) {
							e = p;
							p = p.next;
							psize--;
						}
						else if(cmp(q.val,p.val) >= 0) {
							e = p;
							p = p.next;
							psize--;
						}
						else {
							e = q;
							q = q.next;
							qsize--;
						};
						if(this._valid(tail)) tail.next = e;
						else h = e;
						e.prev = tail;
						tail = e;
					};
					p = q;
				};
				tail.next = null;
				if(nmerges <= 1) break;
				insize <<= 1;
			};
			h.prev = null;
			this.tail = tail;
			return h;
		}
		
		protected function _insertionSortComparable(node : de.polygonal.ds.DLLNode) : de.polygonal.ds.DLLNode {
			var h : de.polygonal.ds.DLLNode = node;
			var n : de.polygonal.ds.DLLNode = h.next;
			while(this._valid(n)) {
				var m : de.polygonal.ds.DLLNode = n.next;
				var p : de.polygonal.ds.DLLNode = n.prev;
				var v : * = n.val;
				if(((function($this:DLL) : de.polygonal.ds.Comparable {
					var $r : de.polygonal.ds.Comparable;
					var $t : * = p.val;
					if(Std._is($t,de.polygonal.ds.Comparable)) (($t) as de.polygonal.ds.Comparable);
					else throw "Class cast error";
					$r = $t;
					return $r;
				}(this))).compare(v) < 0) {
					var i : de.polygonal.ds.DLLNode = p;
					while(i.hasPrev()) if(((function($this:DLL) : de.polygonal.ds.Comparable {
						var $r2 : de.polygonal.ds.Comparable;
						var $t : * = i.prev.val;
						if(Std._is($t,de.polygonal.ds.Comparable)) (($t) as de.polygonal.ds.Comparable);
						else throw "Class cast error";
						$r2 = $t;
						return $r2;
					}(this))).compare(v) < 0) i = i.prev;
					else break;
					if(this._valid(m)) {
						p.next = m;
						m.prev = p;
					}
					else {
						p.next = null;
						this.tail = p;
					};
					if(i == h) {
						n.prev = null;
						n.next = i;
						i.prev = n;
						h = n;
					}
					else {
						n.prev = i.prev;
						i.prev.next = n;
						n.next = i;
						i.prev = n;
					}
				};
				n = m;
			};
			return h;
		}
		
		protected function _insertionSort(node : de.polygonal.ds.DLLNode,cmp : Function) : de.polygonal.ds.DLLNode {
			var h : de.polygonal.ds.DLLNode = node;
			var n : de.polygonal.ds.DLLNode = h.next;
			while(this._valid(n)) {
				var m : de.polygonal.ds.DLLNode = n.next;
				var p : de.polygonal.ds.DLLNode = n.prev;
				var v : * = n.val;
				if(cmp(v,p.val) < 0) {
					var i : de.polygonal.ds.DLLNode = p;
					while(i.hasPrev()) if(cmp(v,i.prev.val) < 0) i = i.prev;
					else break;
					if(this._valid(m)) {
						p.next = m;
						m.prev = p;
					}
					else {
						p.next = null;
						this.tail = p;
					};
					if(i == h) {
						n.prev = null;
						n.next = i;
						i.prev = n;
						h = n;
					}
					else {
						n.prev = i.prev;
						i.prev.next = n;
						n.next = i;
						i.prev = n;
					}
				};
				n = m;
			};
			return h;
		}
		
		protected function _valid(node : de.polygonal.ds.DLLNode) : Boolean {
			return node != null;
		}
		
		protected function _getNode(x : *) : de.polygonal.ds.DLLNode {
			if(this._reservedSize == 0 || this._poolSize == 0) return new de.polygonal.ds.DLLNode(x,this);
			else {
				var n : de.polygonal.ds.DLLNode = this._headPool;
				this._headPool = this._headPool.next;
				this._poolSize--;
				n.next = null;
				n.val = x;
				return n;
			};
			return null;
		}
		
		protected function _putNode(x : de.polygonal.ds.DLLNode) : * {
			var val : * = x.val;
			if(this._reservedSize > 0 && this._poolSize < this._reservedSize) {
				this._tailPool = this._tailPool.next = x;
				x.val = null;
				this._poolSize++;
			}
			else this.__list(x,null);
			return val;
		}
		
		protected function __insertAfter(f : *,x : de.polygonal.ds.DLLNode) : void {
			f._insertAfter(x);
		}
		
		protected function __insertBefore(f : *,x : de.polygonal.ds.DLLNode) : void {
			f._insertBefore(x);
		}
		
		protected function __unlink(f : *) : void {
			f._unlink();
		}
		
		protected function __list(f : *,x : de.polygonal.ds.DLL) : void {
			f._list = x;
		}
		
	}
}
