package haxe {
	import flash.Boot;
	public class Int64 {
		public function Int64(high : int = 0,low : int = 0) : void { if( !flash.Boot.skip_constructor ) {
			this.high = int(high);
			this.low = low;
		}}
		
		public var high : int;
		public var low : int;
		public function toString() : String {
			if((this.high | this.low) == 0) return "0";
			var str : String = "";
			var neg : Boolean = false;
			var i : haxe.Int64 = this;
			if(haxe.Int64.isNeg(i)) {
				neg = true;
				i = haxe.Int64.neg(i);
			};
			var ten : haxe.Int64 = haxe.Int64.ofInt(10);
			while(!haxe.Int64.isZero(i)) {
				var r : * = haxe.Int64.divMod(i,ten);
				str = r.modulus.low + str;
				i = r.quotient;
			};
			if(neg) str = "-" + str;
			return str;
		}
		
		static protected function ushr32(v : int,n : int) : int {
			return v >>> n;
		}
		
		static public function make(high : int,low : int) : haxe.Int64 {
			return new haxe.Int64(high,low);
		}
		
		static public function ofInt(x : int) : haxe.Int64 {
			return new haxe.Int64(x >> 31,x);
		}
		
		static public function toInt(x : haxe.Int64) : int {
			if(x.high != 0) {
				if(x.high < 0) return -haxe.Int64.toInt(haxe.Int64.neg(x));
				throw "Overflow";
			};
			return x.low;
		}
		
		static public function getLow(x : haxe.Int64) : int {
			return x.low;
		}
		
		static public function getHigh(x : haxe.Int64) : int {
			return x.high;
		}
		
		static public function add(a : haxe.Int64,b : haxe.Int64) : haxe.Int64 {
			var high : int = a.high + b.high;
			var low : int = a.low + b.low;
			if(haxe.Int64.uicompare(low,a.low) < 0) high++;
			return new haxe.Int64(high,low);
		}
		
		static public function sub(a : haxe.Int64,b : haxe.Int64) : haxe.Int64 {
			var high : int = a.high - b.high;
			var low : int = a.low - b.low;
			if(haxe.Int64.uicompare(a.low,b.low) < 0) high--;
			return new haxe.Int64(high,low);
		}
		
		static public function mul(a : haxe.Int64,b : haxe.Int64) : haxe.Int64 {
			var mask : int = 65535;
			var al : int = a.low & mask;
			var ah : int = haxe.Int64.ushr32(a.low,16);
			var bl : int = b.low & mask;
			var bh : int = haxe.Int64.ushr32(b.low,16);
			var p00 : int = al * bl;
			var p10 : int = ah * bl;
			var p01 : int = al * bh;
			var p11 : int = ah * bh;
			var low : int = p00;
			var high : int;
			{
				var i : int = p11 + haxe.Int64.ushr32(p01,16) + haxe.Int64.ushr32(p10,16);
				high = i;
			};
			p01 = p01 << 16;
			low = low + p01;
			if(haxe.Int64.uicompare(low,p01) < 0) high = high + 1;
			p10 = p10 << 16;
			low = low + p10;
			if(haxe.Int64.uicompare(low,p10) < 0) high = high + 1;
			high = high + a.low * b.high;
			high = high + a.high * b.low;
			return new haxe.Int64(high,low);
		}
		
		static protected function divMod(modulus : haxe.Int64,divisor : haxe.Int64) : * {
			var quotient : haxe.Int64 = new haxe.Int64(0,0);
			var mask : haxe.Int64 = new haxe.Int64(0,1);
			divisor = new haxe.Int64(divisor.high,divisor.low);
			while(divisor.high >= 0) {
				var cmp : int = haxe.Int64.ucompare(divisor,modulus);
				{
					var i : int = divisor.high << 1 | haxe.Int64.ushr32(divisor.low,31);
					divisor.high = i;
				};
				divisor.low = divisor.low << 1;
				{
					var i1 : int = mask.high << 1 | haxe.Int64.ushr32(mask.low,31);
					mask.high = i1;
				};
				mask.low = mask.low << 1;
				if(cmp >= 0) break;
			};
			while((mask.low | mask.high) != 0) {
				if(haxe.Int64.ucompare(modulus,divisor) >= 0) {
					quotient.high = (quotient.high | mask.high);
					quotient.low = (quotient.low | mask.low);
					modulus = haxe.Int64.sub(modulus,divisor);
				};
				{
					var i2 : int = haxe.Int64.ushr32(mask.low,1) | mask.high << 31;
					mask.low = i2;
				};
				mask.high = haxe.Int64.ushr32(mask.high,1);
				{
					var i3 : int = haxe.Int64.ushr32(divisor.low,1) | divisor.high << 31;
					divisor.low = i3;
				};
				divisor.high = haxe.Int64.ushr32(divisor.high,1);
			};
			return { quotient : quotient, modulus : modulus}
		}
		
		static public function div(a : haxe.Int64,b : haxe.Int64) : haxe.Int64 {
			if(b.high == 0) {
				var _g : int = b.low;
				switch(_g) {
				case 0:
				throw "divide by zero";
				break;
				case 1:
				return new haxe.Int64(a.high,a.low);
				break;
				}
			};
			var sign : Boolean = (a.high < 0 || b.high < 0) && !(a.high < 0 && b.high < 0);
			if(a.high < 0) a = haxe.Int64.neg(a);
			if(b.high < 0) b = haxe.Int64.neg(b);
			var q : haxe.Int64 = haxe.Int64.divMod(a,b).quotient;
			if(sign) return haxe.Int64.neg(q);
			else return q;
			return null;
		}
		
		static public function mod(a : haxe.Int64,b : haxe.Int64) : haxe.Int64 {
			if(b.high == 0) {
				var _g : int = b.low;
				switch(_g) {
				case 0:
				throw "modulus by zero";
				break;
				case 1:
				return haxe.Int64.ofInt(0);
				break;
				}
			};
			var sign : Boolean = a.high < 0;
			if(a.high < 0) a = haxe.Int64.neg(a);
			if(b.high < 0) b = haxe.Int64.neg(b);
			var m : haxe.Int64 = haxe.Int64.divMod(a,b).modulus;
			if(sign) return haxe.Int64.neg(m);
			else return m;
			return null;
		}
		
		static public function shl(a : haxe.Int64,b : int) : haxe.Int64 {
			if((b & 63) == 0) return a;
			else if((b & 63) < 32) return new haxe.Int64(a.high << b | haxe.Int64.ushr32(a.low,32 - (b & 63)),a.low << b);
			else return new haxe.Int64(a.low << b - 32,0);
			return null;
		}
		
		static public function shr(a : haxe.Int64,b : int) : haxe.Int64 {
			if((b & 63) == 0) return a;
			else if((b & 63) < 32) return new haxe.Int64(a.high >> b,haxe.Int64.ushr32(a.low,b) | a.high << 32 - (b & 63));
			else return new haxe.Int64(a.high >> 31,a.high >> b - 32);
			return null;
		}
		
		static public function ushr(a : haxe.Int64,b : int) : haxe.Int64 {
			if((b & 63) == 0) return a;
			else if((b & 63) < 32) return new haxe.Int64(haxe.Int64.ushr32(a.high,b),haxe.Int64.ushr32(a.low,b) | a.high << 32 - (b & 63));
			else return new haxe.Int64(0,haxe.Int64.ushr32(a.high,b - 32));
			return null;
		}
		
		static public function and(a : haxe.Int64,b : haxe.Int64) : haxe.Int64 {
			return new haxe.Int64(a.high & b.high,a.low & b.low);
		}
		
		static public function or(a : haxe.Int64,b : haxe.Int64) : haxe.Int64 {
			return new haxe.Int64(a.high | b.high,a.low | b.low);
		}
		
		static public function xor(a : haxe.Int64,b : haxe.Int64) : haxe.Int64 {
			return new haxe.Int64(a.high ^ b.high,a.low ^ b.low);
		}
		
		static public function neg(a : haxe.Int64) : haxe.Int64 {
			var high : int = ~a.high;
			var low : int = -a.low;
			if(low == 0) high++;
			return new haxe.Int64(high,low);
		}
		
		static public function isNeg(a : haxe.Int64) : Boolean {
			return a.high < 0;
		}
		
		static public function isZero(a : haxe.Int64) : Boolean {
			return (a.high | a.low) == 0;
		}
		
		static protected function uicompare(a : int,b : int) : int {
			if(a < 0) {
				if(b < 0) return ~b - ~a;
				else return 1;
			}
			else if(b < 0) return -1;
			else return a - b;
			return 0;
		}
		
		static public function compare(a : haxe.Int64,b : haxe.Int64) : int {
			var v : int = a.high - b.high;
			if(v != 0) return v;
			else return haxe.Int64.uicompare(a.low,b.low);
			return 0;
		}
		
		static public function ucompare(a : haxe.Int64,b : haxe.Int64) : int {
			var v : int = haxe.Int64.uicompare(a.high,b.high);
			if(v != 0) return v;
			else return haxe.Int64.uicompare(a.low,b.low);
			return 0;
		}
		
		static public function toStr(a : haxe.Int64) : String {
			return a.toString();
		}
		
	}
}
